-- ------------------------------------------------------------------------------ --
--                                TradeSkillMaster                                --
--                          https://tradeskillmaster.com                          --
--    All Rights Reserved - Detailed license information included with addon.     --
-- ------------------------------------------------------------------------------ --

local TSM = select(2, ...) ---@type TSM
local Groups = TSM:NewPackage("Groups") ---@type AddonPackage
local Database = TSM.LibTSMUtil:Include("Database")
local TempTable = TSM.LibTSMUtil:Include("BaseType.TempTable")
local Table = TSM.LibTSMUtil:Include("Lua.Table")
local Operation = TSM.LibTSMTypes:Include("Operation")
local Group = TSM.LibTSMTypes:Include("Group")
local GroupExport = TSM.LibTSMTypes:IncludeClassType("GroupExport")
local ClientInfo = TSM.LibTSMWoW:Include("Util.ClientInfo")
local private = {
	db = nil,
	groupListCache = {},
	settings = nil,
}
local EXPORT_IGNORED_OPERATION_TYPES = {
	"Mailing",
}
local EXPORT_CUSTOM_STRING_SETTINGS = {
	Auctioning = {
		postCap = true,
		keepQuantity = true,
		maxExpires = true,
		undercut = true,
		minPrice = true,
		maxPrice = true,
		normalPrice = true,
		cancelRepostThreshold = true,
		stackSize = not ClientInfo.IsRetail() and true or nil,
	},
	Crafting = {
		minRestock = true,
		maxRestock = true,
		minProfit = true,
		craftPriceMethod = true,
	},
	Shopping = {
		restockQuantity = true,
		maxPrice = true,
	},
	Sniper = {
		belowPrice = true,
	},
	Vendoring = {
		vsMarketValue = true,
		vsMaxMarketValue = true,
		vsDestroyValue = true,
		vsMaxDestroyValue = true,
	},
	Warehousing = {},
}



-- ============================================================================
-- New Modules Functions
-- ============================================================================

function Groups.OnInitialize(settingsDB)
	private.settings = settingsDB:NewView()
		:AddKey("profile", "userData", "groups")
		:AddKey("profile", "userData", "items")
	private.db = Database.NewSchema("GROUPS")
		:AddStringField("groupPath")
		:AddStringField("orderStr")
		:AddBooleanField("hasAuctioningOperation")
		:AddBooleanField("hasCraftingOperation")
		:AddBooleanField("hasMailingOperation")
		:AddBooleanField("hasShoppingOperation")
		:AddBooleanField("hasSniperOperation")
		:AddBooleanField("hasVendoringOperation")
		:AddBooleanField("hasWarehousingOperation")
		:AddBooleanField("hasAnyAuctioningOperation")
		:AddBooleanField("hasAnyCraftingOperation")
		:AddBooleanField("hasAnyMailingOperation")
		:AddBooleanField("hasAnyShoppingOperation")
		:AddBooleanField("hasAnySniperOperation")
		:AddBooleanField("hasAnyVendoringOperation")
		:AddBooleanField("hasAnyWarehousingOperation")
		:AddIndex("groupPath")
		:Commit()
	local operationInfo = TempTable.Acquire()
	for _, typeName in Operation.TypeIterator() do
		operationInfo[typeName] = Operation.GetMaxNumber(typeName)
	end
	assert(Table.Count(operationInfo) == 7)
	Group.SetOperationInfo(operationInfo, Operation.Exists)
	TempTable.Release(operationInfo)
	Groups.RebuildDatabase()
end

function Groups.RebuildDatabase()
	Group.Load(private.settings.groups, private.settings.items)
	wipe(private.groupListCache)
	Groups.RebuildDB()
end

function Groups.CreateQuery()
	return private.db:NewQuery()
		:OrderBy("orderStr", true)
end

function Groups.RebuildDB()
	private.db:TruncateAndBulkInsertStart()
	for groupPath, group in Group.Iterator() do
		local orderStr = Group.GetSortableString(groupPath)
		local hasAuctioningOperation, hasAnyAuctioningOperation = private.HasOperation(group, "Auctioning")
		local hasCraftingOperation, hasAnyCraftingOperation = private.HasOperation(group, "Crafting")
		local hasMailingOperation, hasAnyMailingOperation = private.HasOperation(group, "Mailing")
		local hasShoppingOperation, hasAnyShoppingOperation = private.HasOperation(group, "Shopping")
		local hasSniperOperation, hasAnySniperOperation = private.HasOperation(group, "Sniper")
		local hasVendoringOperation, hasAnyVendoringOperation = private.HasOperation(group, "Vendoring")
		local hasWarehousingOperation, hasAnyWarehousingOperation = private.HasOperation(group, "Warehousing")
		private.db:BulkInsertNewRow(groupPath, orderStr, hasAuctioningOperation, hasCraftingOperation, hasMailingOperation, hasShoppingOperation, hasSniperOperation, hasVendoringOperation, hasWarehousingOperation, hasAnyAuctioningOperation, hasAnyCraftingOperation, hasAnyMailingOperation, hasAnyShoppingOperation, hasAnySniperOperation, hasAnyVendoringOperation, hasAnyWarehousingOperation)
	end
	private.db:BulkInsertEnd()
	wipe(private.groupListCache)
end

function Groups.Move(groupPath, newGroupPath)
	Group.SetItemDBQueriesPaused(true)
	Group.Move(groupPath, newGroupPath)
	Groups.RebuildDB()
	Group.SetItemDBQueriesPaused(false)
end

function Groups.Delete(groupPath)
	Group.SetItemDBQueriesPaused(true)
	Group.Delete(groupPath)
	Groups.RebuildDB()
	Group.SetItemDBQueriesPaused(false)
end

function Groups.GroupIterator()
	if #private.groupListCache == 0 then
		Groups.CreateQuery()
			:Select("groupPath")
			:NotEqual("groupPath", Group.GetRootPath())
			:AsTable(private.groupListCache)
			:Release()
	end
	return ipairs(private.groupListCache)
end

function Groups.GroupIteratorByOperation(moduleName, operationName, overrideOnly)
	local query = private.db:NewQuery()
		:Select("groupPath")
		:Equal("hasAny"..moduleName.."Operation", true)
		:OrderBy("orderStr", true)
	local result = TempTable.Acquire()
	for _, groupPath in query:Iterator() do
		if (not overrideOnly or Group.HasOperationOverride(groupPath, moduleName)) and Group.HasOperation(groupPath, moduleName, operationName) then
			tinsert(result, groupPath)
		end
	end
	query:Release()
	return TempTable.Iterator(result)
end

function Groups.OperationIterator(groupPath, moduleName)
	local result = TempTable.Acquire()
	for _, operationName in Group.OperationIterator(groupPath, moduleName) do
		Operation.UpdateFromRelationships(moduleName, operationName)
		if not Operation.IsIgnored(moduleName, operationName) then
			tinsert(result, operationName)
			tinsert(result, Operation.GetSettings(moduleName, operationName))
		end
	end
	return TempTable.Iterator(result, 2)
end

function Groups.CreateExportObject()
	return GroupExport.New(EXPORT_IGNORED_OPERATION_TYPES, EXPORT_CUSTOM_STRING_SETTINGS)
end



-- ============================================================================
-- Private Helper Functions
-- ============================================================================

function private.HasOperation(group, typeName)
	local hasNonIgnored, hasAny = false, false
	for _, operationName in ipairs(group[typeName]) do
		if not Operation.IsIgnored(typeName, operationName) then
			hasNonIgnored = true
			hasAny = true
			break
		else
			hasAny = true
		end
	end
	return hasNonIgnored, hasAny
end
