-- ------------------------------------------------------------------------------ --
--                                TradeSkillMaster                                --
--                          https://tradeskillmaster.com                          --
--    All Rights Reserved - Detailed license information included with addon.     --
-- ------------------------------------------------------------------------------ --

local TSM = select(2, ...) ---@type TSM
local Shopping = TSM.Operations:NewPackage("Shopping") ---@type AddonPackage
local L = TSM.Include("Locale").GetTable()
local Table = TSM.LibTSMUtil:Include("Lua.Table")
local Operation = TSM.LibTSMTypes:Include("Operation")
local Group = TSM.LibTSMTypes:Include("Group")
local CustomString = TSM.LibTSMTypes:Include("CustomString")
local CustomPrice = TSM.Include("Service.CustomPrice")
local BagTracking = TSM.LibTSMService:Include("Inventory.BagTracking")
local GuildTracking = TSM.Include("Service.GuildTracking")
local AltTracking = TSM.Include("Service.AltTracking")
local Auction = TSM.LibTSMService:Include("Auction")
local Mail = TSM.LibTSMService:Include("Mail")
local private = {
	settingsDB = nil,
	settings = nil,
	connectedCharacterCache = {},
}
local MIN_RESTOCK_VALUE = 0
local MAX_RESTOCK_VALUE = 50000



-- ============================================================================
-- Module Functions
-- ============================================================================

function Shopping.OnInitialize(settingsDB)
	private.settingsDB = settingsDB
	private.settings = settingsDB:NewView()
		:AddKey("global", "coreOptions", "regionWide")
	local operationType = Operation.NewType("Shopping", L["Shopping"], 1)
		:SetDescriptionFunc(private.GetDescription)
		:AddStringSetting("restockQuantity", "0")
		:AddStringSetting("maxPrice", "dbmarket")
		:AddBooleanSetting("showAboveMaxPrice", true)
		:AddTableSetting("restockSources", { alts = false, auctions = false, bank = false, guild = false })
	TSM.Operations.RegisterType(operationType)
end

function Shopping.GetRestockRange()
	return MIN_RESTOCK_VALUE, MAX_RESTOCK_VALUE
end

function Shopping.GetMaxPrice(itemString)
	local operationSettings = private.GetOperationSettings(itemString)
	if not operationSettings then
		return
	end
	local value = CustomString.GetValue(operationSettings.maxPrice, itemString)
	return value
end

function Shopping.ShouldShowAboveMaxPrice(itemString)
	local operationSettings = private.GetOperationSettings(itemString)
	if not operationSettings then
		return
	end
	return operationSettings.showAboveMaxPrice
end

function Shopping.IsFiltered(itemString, itemBuyout)
	local operationSettings = private.GetOperationSettings(itemString)
	if not operationSettings then
		return true
	end
	if operationSettings.showAboveMaxPrice then
		return false
	end
	local maxPrice = CustomString.GetValue(operationSettings.maxPrice, itemString)
	if itemBuyout > (maxPrice or 0) then
		return true, true
	end
	return false
end

function Shopping.ValidateAndGetRestockQuantity(itemString)
	local operationSettings = private.GetOperationSettings(itemString)
	if not operationSettings then
		return false, nil
	end
	local isValid, err = CustomPrice.Validate(operationSettings.maxPrice)
	if not isValid then
		return false, err
	end
	local restockQuantity = nil
	restockQuantity, err = CustomPrice.GetValue(operationSettings.restockQuantity, itemString, true)
	if not restockQuantity then
		return false, err
	elseif restockQuantity < MIN_RESTOCK_VALUE or restockQuantity > MAX_RESTOCK_VALUE then
		return false, format(L["Your restock quantity is invalid. It must be between %d and %s."], MIN_RESTOCK_VALUE, MAX_RESTOCK_VALUE)
	end
	local maxQuantity = nil
	if restockQuantity > 0 then
		local numHave = private.GetInventoryNum(itemString, operationSettings)
		if numHave >= restockQuantity then
			return false, nil
		end
		maxQuantity = restockQuantity - numHave
	end
	if not operationSettings.showAboveMaxPrice and not CustomString.GetValue(operationSettings.maxPrice, itemString) then
		-- we're not showing auctions above the max price and the max price isn't valid for this item, so skip it
		return false, nil
	end
	return true, maxQuantity
end



-- ============================================================================
-- Private Helper Functions
-- ============================================================================

function private.GetDescription(operationSettings)
	if operationSettings.showAboveMaxPrice then
		return format(L["Shopping for auctions including those above the max price."])
	else
		return format(L["Shopping for auctions with a max price set."])
	end
end

function private.GetOperationSettings(itemString)
	itemString = Group.TranslateItemString(itemString)
	local operationName, operationSettings = TSM.Operations.GetFirstOperationByItem("Shopping", itemString)
	if not operationName then
		return
	end
	return operationSettings
end

function private.GetInventoryNum(itemString, operationSettings)
	-- Check the total inventory as an optimization
	if (CustomString.GetSourceValue("NumInventory", itemString) or 0) == 0 then
		return 0
	end

	local numHave = Mail.GetQuantity(itemString)
	if operationSettings.restockSources.bank then
		numHave = numHave + BagTracking.GetTotalQuantity(itemString)
	else
		numHave = numHave + BagTracking.GetBagQuantity(itemString)
	end
	if operationSettings.restockSources.guild then
		numHave = numHave + GuildTracking.GetQuantity(itemString)
	end
	if operationSettings.restockSources.alts or operationSettings.restockSources.auctions then
		local numAuctions = Auction.GetQuantity(itemString)
		local numAlts = 0
		if private.connectedCharacterCache.time ~= GetTime() then
			wipe(private.connectedCharacterCache)
			for _, factionrealm in private.settingsDB:AccessibleRealmIterator("factionrealm", not private.settings.regionWide) do
				for _, character in private.settingsDB:AccessibleCharacterIterator(nil, factionrealm, true) do
					tinsert(private.connectedCharacterCache, factionrealm)
					tinsert(private.connectedCharacterCache, character)
				end
			end
			private.connectedCharacterCache.time = GetTime()
		end
		for _, factionrealm, character in Table.StrideIterator(private.connectedCharacterCache, 2) do
			numAlts = numAlts + AltTracking.GetBagQuantity(itemString, character, factionrealm)
			numAlts = numAlts + AltTracking.GetBankQuantity(itemString, character, factionrealm)
			numAlts = numAlts + AltTracking.GetReagentBankQuantity(itemString, character, factionrealm)
			numAlts = numAlts + AltTracking.GetMailQuantity(itemString, character, factionrealm)
			numAuctions = numAuctions + AltTracking.GetAuctionQuantity(itemString, character, factionrealm)
		end
		if operationSettings.restockSources.alts then
			numHave = numHave + numAlts
		end
		if operationSettings.restockSources.auctions then
			numHave = numHave + numAuctions
		end
	end
	return numHave
end
