-- ------------------------------------------------------------------------------ --
--                                TradeSkillMaster                                --
--                          https://tradeskillmaster.com                          --
--    All Rights Reserved - Detailed license information included with addon.     --
-- ------------------------------------------------------------------------------ --

local LibTSMTypes = select(2, ...).LibTSMTypes
local GoldLog = LibTSMTypes:DefineClassType("GoldLog")
local NamedTupleList = LibTSMTypes:From("LibTSMUtil"):IncludeClassType("NamedTupleList")
local CSV = LibTSMTypes:From("LibTSMUtil"):Include("Format.CSV")
local BinarySearch = LibTSMTypes:From("LibTSMUtil"):Include("Util.BinarySearch")
local private = {}
local CSV_KEYS = { "minute", "copper" }
local COPPER_PER_GOLD = 100 * 100
local MAX_COPPER_VALUE = 10 * 1000 * 1000 * COPPER_PER_GOLD
local ERRONEOUS_ZERO_THRESHOLD = 5 * 1000 * COPPER_PER_GOLD



-- ============================================================================
-- Static Class Functions
-- ============================================================================

---Loads a gold log from the serialized string.
---@param data string The serialized string
---@return GoldLog?
function GoldLog.__static.Load(data)
	local list = private.Decode(data)
	if not list then
		return nil
	end
	return GoldLog(list)
end

---Creates a new empty gold log.
---@return GoldLog
function GoldLog.__static.New()
	return GoldLog(private.CreateList())
end



-- ============================================================================
-- Meta Class Methods
-- ============================================================================

function GoldLog.__private:__init(list)
	self._list = list ---@type NamedTupleList
end



-- ============================================================================
-- Public Class Methods
-- ============================================================================

-- Clean up any erroneous 0 entries, entries which are too high, and duplicate entries
function GoldLog:Clean()
	local didChange = true
	while didChange do
		didChange = false
		for i = self._list:GetNumRows() - 1, 2, -1 do
			local prevValue = self._list:GetRowField(i - 1, "copper")
			local value = self._list:GetRowField(i, "copper")
			local nextValue = self._list:GetRowField(i + 1, "copper")
			if prevValue > ERRONEOUS_ZERO_THRESHOLD and value == 0 and nextValue > ERRONEOUS_ZERO_THRESHOLD then
				-- This is likely an erroneous 0 value
				didChange = true
				self._list:RemoveRow(i)
			end
		end
		for i = self._list:GetNumRows(), 2, -1 do
			local prevValue = self._list:GetRowField(i - 1, "copper")
			local value = self._list:GetRowField(i, "copper")
			if prevValue == value or value > MAX_COPPER_VALUE then
				-- This is either a duplicate or invalid value
				didChange = true
				self._list:RemoveRow(i)
			end
		end
	end
end

---Gets the value at the specified minute.
---@param minute number The minute to get the value at
---@return number
function GoldLog:GetValue(minute)
	if (self:GetStartMinute() or math.huge) > minute then
		-- Specified time is before we had any data
		return 0
	end
	local index, insertIndex = BinarySearch.Raw(self._list:GetNumRows(), minute, private.GetEntryMinute, self._list)
	-- If we didn't find an exact match, the index is the previous one (compared to the insert index)
	-- as that point's gold value is true up until the next point
	index = index or (insertIndex - 1)
	return self._list:GetRowField(index, "copper")
end

---Gets the first minute included in the log.
---@return number?
function GoldLog:GetStartMinute()
	return self._list:GetNumRows() > 0 and self._list:GetRowField(1, "minute") or nil
end

---Gets the last minute included in the log.
---@return number?
function GoldLog:GetEndMinute()
	local numRows = self._list:GetNumRows()
	return numRows > 0 and self._list:GetRowField(numRows, "minute") or nil
end

---Appends a new entry to the gold log, combining it with the last one as applicable.
---@param minute number The minute
---@param copper number The copper value
function GoldLog:Append(minute, copper)
	local prevRowIndex = self._list:GetNumRows()
	if prevRowIndex > 0 and copper == self._list:GetRowField(prevRowIndex, "copper") then
		-- Amount of gold hasn't changed, so nothing to do
	elseif prevRowIndex > 0 and minute == self._list:GetRowField(prevRowIndex, "minute") then
		-- Gold has changed and the previous record is for the current minute so just modify it
		self._list:SetRowField(prevRowIndex, "copper", copper)
	else
		-- Amount of gold changed and we're in a new minute, so insert a new record
		while prevRowIndex > 0 and self._list:GetRowField(prevRowIndex, "minute") > minute - 1 do
			-- The clock may have changed - just delete everything that's too recent
			self._list:RemoveRow(prevRowIndex)
			prevRowIndex = self._list:GetNumRows()
		end
		self._list:InsertRow(minute, copper)
	end
end

---Serializes the gold log so it can be saved.
---@return string
function GoldLog:Serialize()
	local context = CSV.EncodeStart(CSV_KEYS)
	for _, minute, copper in self._list:Iterator() do
		CSV.EncodeAddRowDataRaw(context, minute, copper)
	end
	return CSV.EncodeEnd(context)
end



-- ============================================================================
-- Private Helper Functions
-- ============================================================================

function private.CreateList()
	return NamedTupleList.New("minute", "copper")
end

function private.Decode(data)
	-- Extract to a list
	local decodeContext = CSV.DecodeStart(data, CSV_KEYS)
	if not decodeContext then
		return nil
	end
	local list = private.CreateList()
	for minute, copper in CSV.DecodeIterator(decodeContext) do
		list:InsertRow(tonumber(minute), tonumber(copper))
	end
	CSV.DecodeEnd(decodeContext)
	return list
end

function private.GetEntryMinute(index, list)
	return list:GetRowField(index, "minute")
end
