-- ------------------------------------------------------------------------------ --
--                                TradeSkillMaster                                --
--                          https://tradeskillmaster.com                          --
--    All Rights Reserved - Detailed license information included with addon.     --
-- ------------------------------------------------------------------------------ --

local LibTSMWoW = select(2, ...).LibTSMWoW
local AuctionHouse = LibTSMWoW:Init("API.AuctionHouse")
local Event = LibTSMWoW:Include("Service.Event")
local ClientInfo = LibTSMWoW:Include("Util.ClientInfo")
local private = {
	itemLocation = ItemLocation:CreateEmpty(),
	auctionIdToLink = {},
	auctionIdToItemBuyout = {},
	postHookFuncs = {},
	purchaseHookFuncs = {},
	lastPurchase = {
		link = nil,
		name = nil,
		quantity = nil,
		buyout = nil,
	},
}
local AUCTIONABLE_WOW_TOKEN_ITEM_ID = 122270



-- ============================================================================
-- Module Loading
-- ============================================================================

AuctionHouse:OnModuleLoad(function()
	if ClientInfo.HasFeature(ClientInfo.FEATURES.C_AUCTION_HOUSE) then
		-- Maintain our auctionId maps
		Event.Register("ITEM_SEARCH_RESULTS_UPDATED", private.HandleItemSearchResultsUpdated)
	end
	-- Setup Hooks
	if ClientInfo.HasFeature(ClientInfo.FEATURES.C_AUCTION_HOUSE) then
		hooksecurefunc(C_AuctionHouse, "PostCommodity", private.PostCommodityHook)
		hooksecurefunc(C_AuctionHouse, "PostItem", private.PostItemHook)
		hooksecurefunc(C_AuctionHouse, "PlaceBid", private.PlaceBidHook)
		hooksecurefunc(C_AuctionHouse, "ConfirmCommoditiesPurchase", private.ConfirmCommoditiesPurchaseHook)
	else
		hooksecurefunc("PostAuction", private.PostAuctionHook)
		hooksecurefunc("PlaceAuctionBid", private.PlaceAuctionBidHook)
	end
end)



-- ============================================================================
-- Module Functions
-- ============================================================================

---Returns whether or not an item in the player's bags can be sold on the AH (requires C_AuctionHouse).
---@param bag any
---@param slot any
---@return boolean
function AuctionHouse.IsSellable(bag, slot)
	assert(ClientInfo.HasFeature(ClientInfo.FEATURES.C_AUCTION_HOUSE))
	private.itemLocation:Clear()
	private.itemLocation:SetBagAndSlot(bag, slot)
	return private.itemLocation:IsValid() and C_AuctionHouse.IsSellItemValid(private.itemLocation, false)
end

---Returns whether or not the owned auctions are fully loaded.
---@return boolean
function AuctionHouse.OwnedFullyLoaded()
	return not ClientInfo.HasFeature(ClientInfo.FEATURES.C_AUCTION_HOUSE) or C_AuctionHouse.HasFullOwnedAuctionResults()
end

---Returns whether or not owned auctions are sorted by owner/duration (not available with C_AuctionHouse).
---@return boolean
function AuctionHouse.AreOwnedSortedByOwnerDuration()
	assert(not ClientInfo.HasFeature(ClientInfo.FEATURES.C_AUCTION_HOUSE))
	local numColumns = #AuctionSort.owner_duration
	for i, info in ipairs(AuctionSort.owner_duration) do
		local col, reversed = GetAuctionSort("owner", numColumns - i + 1)
		-- We want to do the opposite order
		reversed = not reversed
		if col ~= info.column or info.reverse ~= reversed then
			return false
		end
	end
	return true
end

---Sorts owned auctions by owner/duration (not available with C_AuctionHouse).
function AuctionHouse.SortOwnedByOwnerDuration()
	AuctionFrame_SetSort("owner", "duration", true)
	SortAuctionApplySort("owner")
end

---Gets the number of owned auctions.
---@return number
function AuctionHouse.GetNumOwned()
	if ClientInfo.HasFeature(ClientInfo.FEATURES.C_AUCTION_HOUSE) then
		return C_AuctionHouse.GetNumOwnedAuctions()
	else
		return GetNumAuctionItems("owner")
	end
end

---Gets info on an owned auction (auctionId of math.huge indicates ignored WoW token).
---@param index number The owned auction index
---@return number? auctionId
---@return string|number|nil itemLinkOrId
---@return string? name
---@return number? texture
---@return number? stackSize
---@return number? quality
---@return number? currentBid
---@return number? buyout
---@return string? highBidder
---@return boolean? isSold
---@return number? duration
function AuctionHouse.GetOwnedInfo(index)
	local auctionId, link, name, texture, stackSize, quality, minBid, buyout, bid, highBidder, saleStatus, duration = nil, nil, nil, nil, nil, nil, nil, nil, nil, nil, nil, nil
	if ClientInfo.HasFeature(ClientInfo.FEATURES.C_AUCTION_HOUSE) then
		local info = C_AuctionHouse.GetOwnedAuctionInfo(index)
		if info.itemKey.itemID == AUCTIONABLE_WOW_TOKEN_ITEM_ID then
			-- We currently just ignore posted tokens, so just return an auctionId as math.huge to indicate this
			return math.huge
		end
		link = info and info.itemLink or info.itemKey.itemID
		bid = info.bidAmount or info.buyoutAmount
		auctionId = info.auctionID
		stackSize = info.quantity
		minBid = bid
		buyout = info.buyoutAmount or 0
		highBidder = info.bidder or ""
		saleStatus = info.status
		duration = info.timeLeftSeconds
	else
		local _
		name, texture, stackSize, quality, _, _, _, minBid, _, buyout, bid, highBidder, _, _, _, saleStatus = GetAuctionItemInfo("owner", index)
		link = name and name ~= "" and GetAuctionItemLink("owner", index)
		if not link then
			return
		end
		duration = GetAuctionItemTimeLeft("owner", index)
		auctionId = index
		highBidder = highBidder or ""
	end
	local currentBid = highBidder ~= "" and bid or minBid
	if saleStatus == 0 then
		if ClientInfo.HasFeature(ClientInfo.FEATURES.C_AUCTION_HOUSE) then
			duration = time() + duration
		end
	elseif saleStatus == 1 then
		if not currentBid and LibTSMWoW.IsRetail() then
			-- Sometimes wow doesn't tell us the current bid on sold auctions on retail
			currentBid = 0
		end
		duration = time() + duration
	else
		error("Invalid sale status: "..tostring(saleStatus))
	end
	if not link then
		return
	end
	return auctionId, link, name, texture, stackSize, quality, currentBid, buyout, highBidder, saleStatus == 1, duration
end

---Register a secure hook function for when an auction is posted.
---@param func fun(duration: number, itemLink: string?, quantity: number?, price: number?) The function to call
function AuctionHouse.SecureHookPost(func)
	tinsert(private.postHookFuncs, func)
end

---Register a secure hook function for when an auction is cancelled.
---@param func fun(auctionId: number) The function to call
function AuctionHouse.SecureHookCancel(func)
	if ClientInfo.HasFeature(ClientInfo.FEATURES.C_AUCTION_HOUSE) then
		hooksecurefunc(C_AuctionHouse, "CancelAuction", func)
	else
		hooksecurefunc("CancelAuction", func)
	end
end

---Register a secure hook function for when an auction is purchased.
---@param func fun(duration: number, itemLink: string?, quantity: number?, price: number?) The function to call
function AuctionHouse.SecureHookPurchase(func)
	tinsert(private.purchaseHookFuncs, func)
end

---Gets info on the last purchase made.
---@return string|number|nil link
---@return string? name
---@return number? quantity
---@return number? buyout
function AuctionHouse.GetLastPurchase()
	return private.lastPurchase.link, private.lastPurchase.name, private.lastPurchase.quantity, private.lastPurchase.buyout
end



-- ============================================================================
-- Private Helper Functions
-- ============================================================================

function private.HandleItemSearchResultsUpdated(_, itemKey)
	wipe(private.auctionIdToLink)
	wipe(private.auctionIdToItemBuyout)
	for i = 1, C_AuctionHouse.GetNumItemSearchResults(itemKey) do
		local info = C_AuctionHouse.GetItemSearchResultInfo(itemKey, i)
		if info.buyoutAmount then
			private.auctionIdToLink[info.auctionID] = info.itemLink
			private.auctionIdToItemBuyout[info.auctionID] = info.buyoutAmount
		end
	end
end

function private.PostCommodityHook(item, duration, quantity, unitPrice)
	private.HandlePostHook(duration, C_Item.GetItemLink(item), quantity, unitPrice)
end

function private.PostItemHook(item, duration, quantity, _, buyout)
	private.HandlePostHook(duration, C_Item.GetItemLink(item), quantity, buyout)
end

function private.PostAuctionHook(_, _, duration)
	private.HandlePostHook(duration)
end

function private.HandlePostHook(duration, itemLink, quantity, price)
	for _, func in ipairs(private.postHookFuncs) do
		func(duration, itemLink, quantity, price)
	end
end

function private.PlaceBidHook(auctionId, bidPlaced)
	local link = private.auctionIdToLink[auctionId]
	local buyout = private.auctionIdToItemBuyout[auctionId]
	if not link or buyout ~= bidPlaced then
		return
	end
	private.HandlePurchaseHook(link, nil, 1, bidPlaced)
end

function private.ConfirmCommoditiesPurchaseHook(itemId, quantity)
	local origQuantity = quantity
	local buyout = 0
	for i = 1, C_AuctionHouse.GetNumCommoditySearchResults(itemId) do
		local info = C_AuctionHouse.GetCommoditySearchResultInfo(itemId, i)
		local resultQuantity = min(quantity, info.quantity - info.numOwnerItems)
		buyout = buyout + resultQuantity * info.unitPrice
		quantity = quantity - resultQuantity
		if quantity == 0 then
			break
		end
	end
	if quantity > 0 then
		return
	end
	private.HandlePurchaseHook(itemId, nil, origQuantity, buyout)
end

function private.PlaceAuctionBidHook(_, index, amountPaid)
	local itemLink = GetAuctionItemLink("list", index)
	local name, _, stackSize, _, _, _, _, _, _, buyout = GetAuctionItemInfo("list", index)
	if amountPaid ~= buyout then
		return
	end
	private.HandlePurchaseHook(itemLink, name, stackSize, buyout)
end

function private.HandlePurchaseHook(link, name, quantity, price)
	private.lastPurchase.link = link
	private.lastPurchase.name = name
	private.lastPurchase.quantity = quantity
	private.lastPurchase.buyout = price
end
